"""
Plots excerpts of the input data with its corresponding metadata.
"""

import numpy as np
import argparse
from pathlib import Path
import matplotlib.pyplot as plt
import seaborn as sns
from utils.io_utils import load_neo, save_plot
from utils.neo_utils import time_slice
from utils.parse import parse_plot_channels, none_or_int, none_or_float

CLI = argparse.ArgumentParser()
CLI.add_argument("--data", nargs='?', type=Path, required=True,
                 help="path to input data in neo format")
CLI.add_argument("--output", nargs='?', type=Path, required=True,
                 help="path of output figure")
CLI.add_argument("--t_start", nargs='?', type=none_or_float, default=0,
                 help="start time in seconds")
CLI.add_argument("--t_stop", nargs='?', type=none_or_float, default=10,
                 help="stop time in seconds")
CLI.add_argument("--channels", nargs='+', type=none_or_int, default=0,
                 help="list of channels to plot")

def plot_traces(asig, channels):
    sns.set(style='ticks', palette="deep", context="notebook")
    fig, ax = plt.subplots()

    offset = np.max(np.abs(asig.as_array()[:,channels]))

    for i, signal in enumerate(asig.as_array()[:,channels].T):
        ax.plot(asig.times, signal + i*offset)

    annotations = [f'{k}: {v}' for k,v in asig.annotations.items()
                               if k not in ['nix_name', 'neo_name']]
    array_annotations = [f'{k}: {v[channels]}'
                        for k,v in asig.array_annotations.items()]

    x_coords = asig.array_annotations['x_coords']
    y_coords = asig.array_annotations['y_coords']
    dim_x, dim_y = np.max(x_coords)+1, np.max(y_coords)+1

    ax.text(ax.get_xlim()[1]*1.05, ax.get_ylim()[0],
            f'ANNOTATIONS FOR CHANNEL(s) {channels} \n'\
          +  '\n ANNOTATIONS:\n' + '\n'.join(annotations) \
          +  '\n\n ARRAY ANNOTATIONS:\n' + '\n'.join(array_annotations) +'\n' \
          + f' t_start: {asig.t_start}; t_stop: {asig.t_stop} \n' \
          + f' dimensions(x,y): {dim_x}, {dim_y}')

    ax.set_xlabel(f'time [{asig.times.units.dimensionality.string}]')
    ax.set_ylabel(f'channels [in {asig.units.dimensionality.string}]')
    ax.set_yticks([i*offset for i in range(len(channels))])
    ax.set_yticklabels(channels)
    return ax


if __name__ == '__main__':
    args, unknown = CLI.parse_known_args()

    asig = load_neo(args.data, 'analogsignal', lazy=True)

    channels = parse_plot_channels(args.channels, args.data)

    asig = time_slice(asig, t_start=args.t_start, t_stop=args.t_stop,
                      lazy=True, channel_indexes=channels)

    ax = plot_traces(asig, channels)
    save_plot(args.output)



"""
Select a region of interest (ROI) by thresholding the intensity signal.
"""

import numpy as np
import matplotlib.pyplot as plt
from skimage import measure
import shapely.geometry as geo
import argparse
import neo
import os
from utils.io_utils import load_neo, write_neo, save_plot
from utils.parse import none_or_str, str_to_bool
from utils.neo_utils import analogsignal_to_imagesequence, imagesequence_to_analogsignal

CLI = argparse.ArgumentParser()
CLI.add_argument("--data", nargs='?', type=str, required=True,
                 help="path to input data in neo format")
CLI.add_argument("--output", nargs='?', type=str, required=True,
                 help="path of output file")
CLI.add_argument("--output_img", nargs='?', type=none_or_str,
                 help="path of output image", default=None)
CLI.add_argument("--intensity_threshold", nargs='?', type=float,
                 help="threshold for mask [0,1]", default=0.5)
CLI.add_argument("--crop_to_selection", nargs='?', type=str_to_bool,
                 help="discard frame outside of ROI", default=True)

def calculate_contour(img, contour_limit):
    # Computing the contour lines...
    vmax = np.max(img)
    vmin = np.min(img)
    limit = vmin + contour_limit * (vmax-vmin)
    contour_fake = measure.find_contours(img, limit)
    # Select the largest contour lines
    max_index = np.argmax([len(c) for c in contour_fake])

    contour = contour_fake[max_index]
    contour = np.flip(contour, 1) # x,y

    def border_intercepts(contour):
        intercept_left = False
        intercept_right = False
        intercept_top = False
        intercept_bot = False

        if not np.all(contour[:,0]):
            intercept_left = True
        if not np.all(contour[:,1]):
            intercept_bot = True
        if not np.all(contour[:,0]-len(img[0])+1):
            intercept_right = True
        if not np.all(contour[:,1]-len(img[1])+1):
            intercept_top = True
        return {'left': intercept_left, 'right': intercept_right,
                'top': intercept_top, 'bottom' : intercept_bot}

    contour_intercepts = border_intercepts(contour)
    if sum(contour_intercepts.values()) > 1:
        includes_corner = True
        del contour_fake[max_index]

        # include secondary connecting contour
        connected = False
        for snd_contour in contour_fake:
            snd_contour = np.flip(snd_contour, 1) # x,y
            if border_intercepts(snd_contour) == contour_intercepts:
                contour = np.append(contour, snd_contour, axis=0)
                connected = True
        # or include corner
        if not connected:
            snd_contour = []
            if contour_intercepts['left'] and contour_intercepts['bottom']:
                snd_contour += [0,0]
            elif contour_intercepts['bottom'] and contour_intercepts['right']:
                snd_contour += [len(img[0])-1,0]
            elif contour_intercepts['left'] and contour_intercepts['bottom']:
                snd_contour += [0,len(img[1]-1)]
            elif contour_intercepts['top'] and contour_intercepts['right']:
                snd_contour += [len(img[0])-1,len(img[1])-1]
            else:
                raise ValueError('The contour is to large, and can not be determined unambigously!')
    else:
        includes_corner = False

    print('Contour includes corner = ', includes_corner)
    return contour


def close_contour(contour, num):
    dx = contour[-1][0] - contour[0][0]
    dy = contour[-1][1] - contour[0][1]
    avg_d = np.mean(np.append(np.abs(np.diff(contour[:,0])),
                              np.abs(np.diff(contour[:,1]))))
    num = int(np.sqrt(dx**2 + dy**2) / avg_d)
    def connect_coords(idx, num):
        connect = [contour[-1][idx], contour[0][idx]]
        return np.linspace(connect[0], connect[1], num)
    contour_connect = np.array([connect_coords(i, num)
                                for i in range(2)]).T
    return np.concatenate((contour, contour_connect))


def contour2mask(contour, dim_x, dim_y):
    mask = np.zeros((dim_y, dim_x), dtype=bool)
    polygon = geo.polygon.Polygon(contour)
    for x in range(dim_x):
        for y in range(dim_y):
            point = geo.Point(x,y)
            if polygon.contains(point):
                mask[y,x] = 1
    return mask


def crop_to_selection(frames):
    frame = frames[0]
    y, x = np.where(np.isfinite(frame))
    x0, x1 = np.min(x), np.max(x)
    y0, y1 = np.min(y), np.max(y)
    return frames[:, y0:y1+1, x0:x1+1]


def plot_roi(img, contour):
    fig, ax = plt.subplots()
    ax.imshow(img, interpolation='nearest', cmap=plt.cm.gray, origin='lower')
    ax.axis('image')
    ax.set_xticks([])
    ax.set_yticks([])
    ax.plot(contour[:,0], contour[:,1], linewidth=2)
    plt.draw()
    return ax


if __name__ == '__main__':
    args, unknown = CLI.parse_known_args()

    block = load_neo(args.data)
    asig = block.segments[0].analogsignals[0]
    imgseq = analogsignal_to_imagesequence(asig)

    # get average image
    imgseq_array = imgseq.as_array()
    dim_t, dim_y, dim_x = imgseq_array.shape
    avg_img = np.mean(imgseq_array, axis=0)

    # calculate mask
    contour = calculate_contour(img=avg_img,
                                contour_limit=args.intensity_threshold)
    contour = close_contour(contour, num=100)
    mask = contour2mask(contour=contour,
                        dim_x=dim_x,
                        dim_y=dim_y)

    # apply mask
    imgseq_array[:, np.bitwise_not(mask)] = np.nan
    if args.crop_to_selection:
        imgseq_array = crop_to_selection(imgseq_array)

    # replace analogsingal
    tmp_blk = neo.Block()
    tmp_seg = neo.Segment()
    tmp_blk.segments.append(tmp_seg)
    tmp_imgseq = imgseq.duplicate_with_new_data(imgseq_array)
    new_asig = imagesequence_to_analogsignal(tmp_imgseq)

    # save data and figure
    new_asig.description += "Border regions with mean intensity below "\
                         + "{args.intensity_threshold} were discarded. "\
                         + "({})".format(os.path.basename(__file__))
    block.segments[0].analogsignals = [new_asig]

    plot_roi(avg_img, contour)
    save_plot(args.output_img)

    write_neo(args.output, block)



import os

# path for generated data
output_path = os.path.join(os.path.expanduser('~'), 'path/to/pipeline/output/folder/')

# optional alternative path for config files
# directory must contain stageXY_<stage-name>/config_<PROFILE>.yaml
# if None uses the pipeline working directory
configs_dir = None



"""
Detect waves by clustering triggers that are close to each other in time and space.
"""

import neo
import numpy as np
import quantities as pq
import argparse
from sklearn.cluster import DBSCAN
from utils.io_utils import load_neo, write_neo
from utils.neo_utils import remove_annotations

CLI = argparse.ArgumentParser()
CLI.add_argument("--data", nargs='?', type=str, required=True,
                 help="path to input data in neo format")
CLI.add_argument("--output", nargs='?', type=str, required=True,
                 help="path of output file")
CLI.add_argument("--metric", nargs='?', type=str, default='euclidean',
                 help="parameter for sklearn.cluster.DBSCAN")
CLI.add_argument("--time_space_ratio", nargs='?', type=float, default=1,
                 help="factor to apply to time values")
CLI.add_argument("--neighbour_distance", nargs='?', type=float, default=30,
                 help="eps parameter in sklearn.cluster.DBSCAN")
CLI.add_argument("--min_samples", nargs='?', type=int, default=10,
                 help="minimum number of trigger times to form a wavefront")
                    
def cluster_triggers(event, metric, neighbour_distance, min_samples,
                     time_space_ratio, sampling_rate):
    up_idx = np.where(event.labels == 'UP')[0]
    # build 3D array of trigger times
    triggers = np.zeros((len(up_idx), 3))
    triggers[:,0] = event.array_annotations['x_coords'][up_idx]
    triggers[:,1] = event.array_annotations['y_coords'][up_idx]
    triggers[:,2] = event.times[up_idx].rescale('s') \
                    * sampling_rate.rescale('Hz') * time_space_ratio

    clustering = DBSCAN(eps=neighbour_distance,
                        min_samples=min_samples,
                        metric=metric)
    clustering.fit(triggers)

    if len(np.unique(clustering.labels_)) < 1:
        raise ValueError("No clusters found, please adapt the parameters!")

    # remove unclassified trigger points (label == -1)
    cluster_idx = np.where(clustering.labels_ != -2)[0]
    if not len(cluster_idx):
        raise ValueError("Clusters couldn't be classified, please adapt the parameters!")

    wave_idx = up_idx[cluster_idx]

    evt = neo.Event(times=event.times[wave_idx],
                    labels=clustering.labels_[cluster_idx].astype(str),
                    name='wavefronts',
                    array_annotations={'channels':event.array_annotations['channels'][wave_idx],
                                       'x_coords':triggers[:,0][cluster_idx].astype(int),
                                       'y_coords':triggers[:,1][cluster_idx].astype(int)},
                    description='transitions from down to up states. '\
                               +'Labels are ids of wavefronts. '
                               +'Annotated with the channel id ("channels") and '\
                               +'its position ("x_coords", "y_coords").',
                    cluster_algorithm='sklearn.cluster.DBSCAN',
                    cluster_eps=neighbour_distance,
                    cluster_metric=metric,
                    cluster_min_samples=min_samples)

    remove_annotations(event, del_keys=['nix_name', 'neo_name'])
    evt.annotations.update(event.annotations)
    return evt

if __name__ == '__main__':
    args, unknown = CLI.parse_known_args()

    block = load_neo(args.data)
    asig = block.segments[0].analogsignals[0]

    evts = block.filter(name='transitions', objects="Event")[0]

    if len(evts):
        wave_evt = cluster_triggers(event=evts,
                                    metric=args.metric,
                                    neighbour_distance=args.neighbour_distance,
                                    min_samples=args.min_samples,
                                    time_space_ratio=args.time_space_ratio,
                                    sampling_rate=asig.sampling_rate)

    else:
        wave_evt = neo.Event(name='wavefronts', 
                             times=np.array([])*pq.s, labels=[])

    block.segments[0].events.append(wave_evt)

    write_neo(args.output, block)



"""
Offsets the signal in all channels by a fixed value.
"""

import argparse
from pathlib import Path
import matplotlib.pyplot as plt
from utils.io_utils import load_neo, write_neo, save_plot
from utils.parse import none_or_int, none_or_float, none_or_str
from utils.neo_utils import time_slice

CLI = argparse.ArgumentParser()
CLI.add_argument("--data", nargs='?', type=Path, required=True,
                 help="path to input data in neo format")
CLI.add_argument("--output", nargs='?', type=Path, required=True,
                 help="path of output file")
CLI.add_argument("--offset", nargs='?', type=none_or_float, default=None,
                 help="offset the signal by some value")
CLI.add_argument("--img_dir", nargs='?', type=none_or_str, default=None,
                 help="path of figure directory")
CLI.add_argument("--img_name", nargs='?', type=str, default='offset_channel0.png',
                 help='example image filename for channel 0')
CLI.add_argument("--plot_channels", nargs='+', type=none_or_int, default=None,
                 help="list of channels to plot")
CLI.add_argument("--plot_tstart", nargs='?', type=none_or_float, default=0,
                 help="plotting start time in seconds")
CLI.add_argument("--plot_tstop",  nargs='?', type=none_or_float, default=10,
                 help="plotting stop time in seconds")

def offset_signal(asig, offset=None):
    if offset is None:
        offset = 0

    new_signal = asig.as_array() + offset

    new_asig = asig.duplicate_with_new_data(new_signal)

    new_asig.array_annotate(**asig.array_annotations)
    new_asig.annotate(offset=offset)
    new_asig.description += f"Offset by {offset_signal} ({__file__}). "

    return new_asig


def plot_signal(asig, new_asig, channel=0, t_start=None, t_stop=None):
    fig, ax = plt.subplots(figsize=(17,8))

    asig = time_slice(asig, t_start=t_start, t_stop=t_stop)
    ax.plot(asig.times, asig.as_array()[:,channel], color='b',
            linewidth=1, label='original signal')

    new_asig = time_slice(new_asig, t_start=t_start, t_stop=t_stop)
    ax.plot(new_asig.times, new_asig.as_array()[:,channel], color='g', 
            linewidth=1, label='offset signal')

    ax.set_xlabel(f'time [{asig.times.dimensionality.string}]')
    ax.set_ylabel('signal')
    plt.legend()
    return None


if __name__ == '__main__':
    args, unknown = CLI.parse_known_args()

    # LOADING
    block = load_neo(args.data)
    asig = block.segments[0].analogsignals[0]

    # PERFORMING METHOD
    new_asig = offset_signal(asig, offset=args.offset)
    block.segments[0].analogsignals[0] = new_asig

    # PLOTTING
    if args.plot_channels[0] is not None:
        if args.img_dir is None:
            args.img_dir = str(args.output.parent)
        for channel in args.plot_channels:
            plot_signal(asig, new_asig, channel=channel,
                        t_start=args.plot_tstart, t_stop=args.plot_tstop)
            output_path = Path(args.img_dir) \
                        / args.img_name.replace('_channel0', f'_channel{channel}')
            save_plot(output_path)

    # SAVING
    write_neo(args.output, block)



"""
Plot snapshots of the input data showing the detected waves.
"""

import os
import numpy as np
import quantities as pq
import argparse
import matplotlib.pyplot as plt
from utils.io_utils import load_neo, save_plot
from utils.neo_utils import analogsignal_to_imagesequence

CLI = argparse.ArgumentParser()
CLI.add_argument("--data", nargs='?', type=str,
                 help="path to input data in neo format")
CLI.add_argument("--output_dir", nargs='?', type=str,
                 help="path to output directory")
CLI.add_argument("--img_name", nargs='?', type=str,
                 help="")
CLI.add_argument("--time_window", nargs='?', type=float, default=0.4,
                 help="size of the plotted window in seconds.")
CLI.add_argument("--colormap", nargs='?', type=str, default='viridis',
                 help="")

def plot_wave(wave_id, waves_event, asig, frames, vec_frames,
              time_window=0.4*pq.s, cmap='virids'):
    idx = np.where(waves_event.labels == str(wave_id))[0]
    x_coords = waves_event.array_annotations['x_coords'][idx]
    y_coords = waves_event.array_annotations['y_coords'][idx]

    t = waves_event.times[idx]
    time_steps = np.unique(t)

    dim_t, dim_y, dim_x = frames.shape
    y_idx, x_idx = np.meshgrid(np.arange(dim_y), np.arange(dim_x), indexing='ij')
    markersize = 50 / max([dim_x, dim_y])
    skip_step = int(min([dim_x, dim_y]) / 50) + 1

    vmin, vmax = np.nanmin(frames), np.nanmax(frames)

    time_window_steps = asig.sampling_rate.rescale('Hz') * time_window.rescale('s')
    half_window = int(time_window_steps.magnitude / 2)

    fig, axes = plt.subplots(nrows=2, ncols=len(time_steps),
                             figsize=(2*len(time_steps), 5), sharey='row')
    if len(time_steps) < 2:
        axes = [[axes[0]], [axes[1]]]

    axes[0][0].set_ylabel(f"Wave {wave_id}", fontsize=20)

    for i, ax in enumerate(axes[0]):
        ax.set_title('{:.3f} '.format(time_steps[i].magnitude) + f'{time_steps[i].dimensionality}')

    for i in idx:
        x = waves_event.array_annotations['x_coords'][i]
        y = waves_event.array_annotations['y_coords'][i]
        t = waves_event.times[i]
        ax_i = np.where(time_steps == t)[0][0]

        channel = waves_event.array_annotations['channels'][i]
        t_i = np.argmax(asig.times >= t)
        i_start = np.max([0, t_i-half_window])
        i_stop = np.min([t_i+half_window, len(asig.times)-1])
        axes[0][ax_i].plot(asig.times[i_start : i_stop],
                           asig.as_array()[i_start : i_stop, channel])
        axes[0][ax_i].axvline(t, color='r')

        axes[1][ax_i].imshow(frames[t_i], origin='lower', cmap=cmap,
                             vmin=vmin, vmax=vmax)
        axes[1][ax_i].plot(x, y, linestyle='None', marker='D',
                           markersize=markersize, color='r')
        axes[1][ax_i].set_axis_off()

        axes[1][ax_i].quiver(x_idx[::skip_step,::skip_step],
                             y_idx[::skip_step,::skip_step],
                             np.real(vec_frames[t_i])[::skip_step,::skip_step],
                             np.imag(vec_frames[t_i])[::skip_step,::skip_step],
                             zorder=5)
    return axes


if __name__ == '__main__':
    args, unknown = CLI.parse_known_args()

    block = load_neo(args.data)

    asig = block.segments[0].analogsignals[0]
    vec_asig = block.filter(name='optical_flow', objects="AnalogSignal")[0]

    frames = analogsignal_to_imagesequence(asig).as_array()
    vec_frames = analogsignal_to_imagesequence(vec_asig).as_array()

    waves_event = block.filter(name='wavefronts', objects="Event")[0]

    cmap = plt.get_cmap(args.colormap)

    for wave_id in np.unique(waves_event.labels):
        if int(wave_id) != -1:  # collection of not-clustered triggers
            ax = plot_wave(wave_id=wave_id,
                           waves_event=waves_event,
                           asig=asig,
                           frames=frames,
                           vec_frames=vec_frames,
                           time_window=args.time_window*pq.s,
                           cmap=cmap)

            output_path = os.path.join(args.output_dir,
                                       args.img_name.replace('id0', f'id{wave_id}'))
            save_plot(output_path)
            plt.close()



"""
Calculate the wave directions per wave and channel,
based on the optical flow at wave trigger times and locations.
"""

import argparse
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from utils.io_utils import load_neo, save_plot
from utils.parse import none_or_str

CLI = argparse.ArgumentParser()
CLI.add_argument("--data", nargs='?', type=str, required=True,
                 help="path to neo object")
CLI.add_argument("--output", nargs='?', type=str, required=True,
                 help="path of output file")
CLI.add_argument("--output_img", nargs='?', type=none_or_str, default=None,
                 help="path of output image file")
CLI.add_argument("--event_name", "--EVENT_NAME", nargs='?', type=str, default='wavefronts',
                 help="name of neo.Event to analyze (must contain waves)")

if __name__ == '__main__':
    args, unknown = CLI.parse_known_args()

    block = load_neo(args.data)
    evt = block.filter(name=args.event_name, objects="Event")[0]
    evt = evt[evt.labels.astype('str') != '-1']

    optical_flow = block.filter(name='optical_flow', objects="AnalogSignal")[0]

    df_dict = {f'{args.event_name}_id': evt.labels,
               'channel_id': evt.array_annotations['channels'],
               'flow_direction_local_x': np.empty(len(evt), dtype=float),
               'flow_direction_local_y': np.empty(len(evt), dtype=float),
               }

    for i, trigger in enumerate(evt):
        t_idx = np.argmax(optical_flow.times >= trigger)
        channel_id = evt.array_annotations['channels'][i],
        direction = optical_flow[t_idx, channel_id]
        df_dict['flow_direction_local_x'][i] = np.real(direction)
        df_dict['flow_direction_local_y'][i] = np.imag(direction)

    df = pd.DataFrame(df_dict)
    df.to_csv(args.output)

    fig, ax = plt.subplots(subplot_kw={'projection': 'polar'})
    ax.hist(np.angle(df.flow_direction_local_x + 1j*df.flow_direction_local_y), 
            bins=36, range=[-np.pi, np.pi])

    if args.output_img is not None:
        save_plot(args.output_img)



"""
Calculate the wave propagation velocity for each wave and channel.
"""

import argparse
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from utils.io_utils import load_neo, save_plot
from utils.parse import none_or_str

CLI = argparse.ArgumentParser()
CLI.add_argument("--data", nargs='?', type=str, required=True,
                 help="path to spatial derivative dataframe")
CLI.add_argument("--output", nargs='?', type=str, required=True,
                 help="path of output file")
CLI.add_argument("--output_img", nargs='?', type=none_or_str, default=None,
                 help="path of output image file")
CLI.add_argument("--event_name", "--EVENT_NAME", nargs='?', type=str, default='wavefronts',
                 help="name of neo.Event to analyze (must contain waves)")

if __name__ == '__main__':
    args, unknown = CLI.parse_known_args()

    df = pd.read_csv(args.data)

    velocity = df.spatial_scale * np.sqrt(1/(df.dt_x**2 + df.dt_y**2))
    velocity[~np.isfinite(velocity)] = np.nan

    velocity_df = pd.DataFrame(velocity, columns=['velocity_local'])
    velocity_df['channel_id'] = df.channel_id
    velocity_df['velocity_local_unit'] = f'{df.spatial_scale_unit[0]}/{df.dt_unit[0]}'
    velocity_df[f'{args.event_name}_id'] = df[f'{args.event_name}_id']

    velocity_df.to_csv(args.output)

    if args.output_img is not None:
        save_plot(args.output_img)



"""
Subsample the input data to a target rate by selecting only every n-th sample point.
"""

import argparse
import quantities as pq
import numpy as np
from pathlib import Path
from utils.io_utils import load_neo, write_neo

CLI = argparse.ArgumentParser()
CLI.add_argument("--data", nargs='?', type=str, required=True,
                 help="path to input data in neo format")
CLI.add_argument("--output", nargs='?', type=str, required=True,
                 help="path of output file")
CLI.add_argument("--target_rate", nargs='?', type=float, required=True,
                 help="rate to subsample to in Hz")

if __name__ == '__main__':
    args, unknown = CLI.parse_known_args()

    block = load_neo(args.data)
    asig = block.segments[0].analogsignals[0]

    subsampling_order = asig.sampling_rate/(args.target_rate*pq.Hz)
    subsampling_order = int(np.round(subsampling_order.rescale('dimensionless')))

    sub_asig = asig.duplicate_with_new_data(asig.as_array()[::subsampling_order])
    sub_asig.sampling_rate = asig.sampling_rate/subsampling_order

    sub_asig.array_annotations = asig.array_annotations
    block.segments[0].analogsignals[0] = sub_asig

    write_neo(args.output, block)



"""
Replace the data signal value with their corresponding Hilbert phase.
"""

import numpy as np
from elephant.signal_processing import hilbert
import argparse
import os
from utils.io_utils import load_neo, write_neo

CLI = argparse.ArgumentParser()
CLI.add_argument("--data", nargs='?', type=str, required=True,
                 help="path to input data in neo format")
CLI.add_argument("--output", nargs='?', type=str, required=True,
                 help="path of output file")

if __name__ == '__main__':
    args, unknown = CLI.parse_known_args()

    block = load_neo(args.data)
    asig = block.segments[0].analogsignals[0]

    phase = np.angle(hilbert(asig).as_array())

    asig = asig.duplicate_with_new_data(phase)
    asig.array_annotations = block.segments[0].analogsignals[0].array_annotations

    asig.description += "Phase signal ({}). "\
                        .format(os.path.basename(__file__))
    block.segments[0].analogsignals[0] = asig

    write_neo(args.output, block)



"""
Detect trigger times (i.e., state transition / local wavefronts onsets) 
by finding local minima preceding a prominent peak in the channel signals.
"""

import neo
import numpy as np
import quantities as pq
from scipy.signal import find_peaks
import argparse
from pathlib import Path
from utils.io_utils import load_neo, write_neo, save_plot
from utils.neo_utils import remove_annotations, time_slice
from utils.parse import none_or_int, none_or_float, none_or_str
import seaborn as sns
import matplotlib.pyplot as plt


CLI = argparse.ArgumentParser()
CLI.add_argument("--data", nargs='?', type=Path, required=True,
                 help="path to input data in neo format")
CLI.add_argument("--output", nargs='?', type=Path, required=True,
                 help="path of output file")
CLI.add_argument("--num_interpolation_points", nargs='?', type=int, default=0,
                 help="number of neighboring points to interpolate")
CLI.add_argument("--minima_persistence", nargs='?', type=float, default=0.200,
                 help="minimum time minima (s)")
CLI.add_argument("--maxima_threshold_fraction", nargs='?', type=float, default=0.5,
                 help="amplitude fraction (in range [0,1]) to set the threshold for detecting local maxima")
CLI.add_argument("--maxima_threshold_window", nargs='?', type=none_or_float, default=None,
                 help="time window (s) to set the threshold for detecting local maxima")
CLI.add_argument("--min_peak_distance", nargs='?', type=float, default=0.200,
                 help="minimum distance between peaks (s)")
CLI.add_argument("--img_dir", nargs='?', type=Path,
                 default=None, help="path of figure directory")
CLI.add_argument("--img_name", nargs='?', type=str,
                 default='minima_channel0.png',
                 help='example image filename for channel 0')
CLI.add_argument("--plot_channels", nargs='+', type=none_or_int, default=None,
                 help="list of channels to plot")
CLI.add_argument("--plot_tstart", nargs='?', type=none_or_float, default=0,
                 help="start time (s)")
CLI.add_argument("--plot_tstop", nargs='?', type=none_or_float, default=10,
                 help="stop time (s)")

def filter_minima_order(signal, mins, order=1):
    filtered_mins = np.array([], dtype=int)

    for midx in mins:
        idx = np.arange(midx+1, midx+order, dtype=int)
        idx = idx[idx < len(signal)]

        if (signal[midx] < signal[idx]).all():
            filtered_mins = np.append(filtered_mins, midx)

    return filtered_mins


def moving_threshold(signal, maxima_threshold_window, maxima_threshold_fraction):
    threshold_signal = []

    sampling_rate = signal.sampling_rate.rescale('Hz').magnitude
    duration = float(signal.t_stop.rescale('s').magnitude - signal.t_start.rescale('s').magnitude)
    if maxima_threshold_window is None or maxima_threshold_window > duration:
        maxima_threshold_window = duration
    window_frame = int(maxima_threshold_window*sampling_rate)

    for channel, channel_signal in enumerate(signal.T):
        if np.isnan(channel_signal).any():
            threshold_func = np.full(np.shape(signal)[0], np.nan)
        else:
            # compute a dynamic threshold function through a sliding window
            # on the signal array
            strides = np.lib.stride_tricks.sliding_window_view(channel_signal, window_frame)
            threshold_func = np.min(strides, axis=1) + maxima_threshold_fraction*np.ptp(strides, axis=1)
            # add elements at the beginning
            threshold_func = np.append(np.ones(window_frame//2)*threshold_func[0], threshold_func)
            threshold_func = np.append(threshold_func, np.ones(len(channel_signal)-len(threshold_func))*threshold_func[-1])
        threshold_signal.append(threshold_func)

    threshold_signal = neo.AnalogSignal(np.array(threshold_signal).T, units=signal.units, sampling_rate=signal.sampling_rate)

    return threshold_signal


def detect_minima(asig, threshold_asig, interpolation_points,
                  min_peak_distance, minima_persistence):
    signal = asig.as_array()
    times = asig.times.rescale('s').magnitude
    sampling_rate = asig.sampling_rate.rescale('Hz').magnitude
    threshold = threshold_asig.as_array()

    min_time_idx = np.array([], dtype=int)
    channel_idx  = np.array([], dtype=int)

    minima_order = int(np.max([minima_persistence*sampling_rate, 1]))
    min_distance = np.max([min_peak_distance*sampling_rate, 1])

    for channel, channel_signal in enumerate(signal.T):
        if np.isnan(channel_signal).any(): continue

        peaks, _ = find_peaks(channel_signal, distance=min_distance, height=threshold.T[channel])
        dmins, _ = find_peaks(-channel_signal)#, distance=min_distance)

        mins = filter_minima_order(channel_signal, dmins, order=minima_order)

        clean_mins = np.array([], dtype=int)

        for i, peak in enumerate(peaks):
            distance_to_peak = times[peak] - times[mins]
            distance_to_peak = distance_to_peak[distance_to_peak > 0]
            if distance_to_peak.size:
                trans_idx = np.argmin(distance_to_peak)
                clean_mins = np.append(clean_mins, mins[trans_idx])

        min_time_idx = np.append(min_time_idx, clean_mins)
        channel_idx = np.append(channel_idx, np.ones(len(clean_mins), dtype=int)*channel)

    # compute local minima times.
    if interpolation_points:
        # parabolic fit on the right branch of local minima
        start_arr = min_time_idx - 1
        start_arr = np.where(start_arr > 0, start_arr, 0)
        stop_arr = start_arr + int(interpolation_points)

        start_arr = np.where(stop_arr < len(signal), start_arr, len(signal)-interpolation_points-1)
        stop_arr = np.where(stop_arr < len(signal), stop_arr, len(signal)-1)

        signal_arr = np.empty((interpolation_points, len(start_arr)))
        signal_arr.fill(np.nan)

        for i, (start, stop, channel_i) in enumerate(zip(start_arr, stop_arr, channel_idx)):
            signal_arr[:,i] = signal[start:stop, channel_i]

        X_temp = range(0, interpolation_points)
        params = np.polyfit(X_temp, signal_arr, 2)

        min_pos = -params[1,:] / (2*params[0,:]) + start_arr
        min_pos = np.where(min_pos > 0, min_pos, 0)
        minimum_times = min_pos/sampling_rate*pq.s
    else:
        minimum_times = asig.times[min_time_idx]

    idx = np.where(minimum_times >= np.max(asig.times))[0]
    minimum_times[idx] = np.max(asig.times)
    ###################################
    sort_idx = np.argsort(minimum_times)

    # save detected minima as transition
    evt = neo.Event(times=minimum_times[sort_idx],
                    labels=['UP'] * len(minimum_times),
                    name='transitions',
                    trigger_detection='minima',
                    num_interpolation_points=interpolation_points,
                    array_annotations={'channels':channel_idx[sort_idx]})

    for key in asig.array_annotations.keys():
        evt_ann = {key : asig.array_annotations[key][channel_idx[sort_idx]]}
        evt.array_annotations.update(evt_ann)

    remove_annotations(asig)
    evt.annotations.update(asig.annotations)

    return evt


def plot_minima(asig, event, threshold_asig, channel, min_peak_distance):
    signal = asig.as_array().T[channel]
    times = asig.times.rescale('s')
    sampling_rate = asig.sampling_rate.rescale('Hz').magnitude
    threshold = threshold_asig.as_array().T[channel]

    peaks, _ = find_peaks(signal, height=threshold,
                          distance=np.max([min_peak_distance*sampling_rate, 1]))  

    # plot figure
    sns.set(style='ticks', palette="deep", context="notebook")
    fig, ax = plt.subplots()

    ax.plot(times, signal, label='signal', color='k')
    ax.plot(times, threshold, label='moving threshold',
            linestyle=':', color='b')

    idx_ch = np.where(event.array_annotations['channels'] == channel)[0]

    ax.plot(times[peaks], signal[peaks], 'x', color='r', label='detected maxima') 
    ax.plot(event.times[idx_ch],
            signal[((event.times[idx_ch]-asig.times[0])*sampling_rate).astype(int)],
            'x', color='g', label='selected minima')

    ax.set_title(f'channel {channel}')
    ax.set_xlabel('time [s]')
    ax.legend()

    return ax


if __name__ == '__main__':
    args, unknown = CLI.parse_known_args()

    block = load_neo(args.data)
    asig = block.segments[0].analogsignals[0]

    threshold_asig = moving_threshold(asig, args.maxima_threshold_window, args.maxima_threshold_fraction)

    transition_event = detect_minima(asig, threshold_asig,
                                     interpolation_points=args.num_interpolation_points,
                                     min_peak_distance=args.min_peak_distance,
                                     minima_persistence=args.minima_persistence)

    block.segments[0].events.append(transition_event)

    write_neo(args.output, block)

    if args.plot_channels[0] is not None:
        for channel in args.plot_channels:
            plot_minima(asig=time_slice(asig, args.plot_tstart, args.plot_tstop),
                        event=time_slice(transition_event, args.plot_tstart, args.plot_tstop),
                        threshold_asig=time_slice(threshold_asig, args.plot_tstart, args.plot_tstop),
                        channel=int(channel),
                        min_peak_distance=args.min_peak_distance)
            output_path = args.img_dir / args.img_name.replace('_channel0', f'_channel{channel}')
            save_plot(output_path)



"""
Merge pandas DataFrames based on the values of selected columns.
"""

import argparse
import pandas as pd
from copy import deepcopy

CLI = argparse.ArgumentParser()
CLI.add_argument("--data", nargs='+', type=str,
                 help="path to input data in neo format")
CLI.add_argument("--output", nargs='?', type=str,
                 help="path of output file")
CLI.add_argument("--output_img", nargs='?', type=str,
                 help="")
CLI.add_argument("--merge_key", nargs='+', type=str,
                 help="")

if __name__ == '__main__':
    args, unknown = CLI.parse_known_args()

    for i, datafile in enumerate(args.data):
        df = pd.read_csv(datafile)
        df.drop(df.columns[df.columns.str.contains('unnamed', case=False)],
                axis=1, inplace=True)
        if i:
            full_df = full_df.merge(df, how='outer', on=args.merge_key)
        else:
            full_df = deepcopy(df)
        del df

    full_df.to_html(args.output_img)

    full_df.to_csv(args.output)



"""
Plot an example signal trace before and after application of some processing steps.
"""

import matplotlib.pyplot as plt
import seaborn as sns
import argparse
import os
from utils.io_utils import load_neo, save_plot
from utils.neo_utils import time_slice
from utils.parse import none_or_float

CLI = argparse.ArgumentParser()
CLI.add_argument("--original_data", nargs='?', type=str, required=True,
                 help="path to input data in neo format")
CLI.add_argument("--data", nargs='?', type=str, required=True,
                 help="path to input data in neo format")
CLI.add_argument("--img_dir", nargs='?', type=str, required=True,
                 help="path of output figure directory")
CLI.add_argument("--img_name", nargs='?', type=str,
                 default='processed_trace_channel0.png',
                 help='example filename for channel 0')
CLI.add_argument("--t_start", nargs='?', type=none_or_float, default=0,
                 help="start time in seconds")
CLI.add_argument("--t_stop", nargs='?', type=none_or_float, default=10,
                 help="stop time in seconds")
CLI.add_argument("--channels", nargs='+', type=int, default=0,
                 help="channel to plot")

def plot_traces(original_asig, processed_asig, channel):
    sns.set(style='ticks', palette="deep", context="notebook")
    fig, ax1 = plt.subplots()
    palette = sns.color_palette()

    ax1.plot(original_asig.times,
            original_asig.as_array()[:,channel],
            color=palette[0])
    ax1.set_ylabel('original signal', color=palette[0])
    ax1.tick_params('y', colors=palette[0])

    ax2 = ax1.twinx()
    ax2.plot(processed_asig.times,
            processed_asig.as_array()[:,channel],
            color=palette[1])
    ax2.set_ylabel('processed signal', color=palette[1])
    ax2.tick_params('y', colors=palette[1])

    ax1.set_title('Channel {}'.format(channel))
    ax1.set_xlabel('time [{}]'.format(original_asig.times.units.dimensionality.string))

    return ax1, ax2


if __name__ == '__main__':
    args, unknown = CLI.parse_known_args()

    orig_asig = load_neo(args.original_data, 'analogsignal', lazy=False)
    orig_asig = time_slice(orig_asig, t_start=args.t_start, t_stop=args.t_stop,
                           lazy=False, channel_indexes=args.channels)

    proc_asig = load_neo(args.data, 'analogsignal', lazy=False)
    proc_asig = time_slice(proc_asig, t_start=args.t_start, t_stop=args.t_stop,
                           lazy=False, channel_indexes=args.channels)

    for channel in args.channels:
        plot_traces(orig_asig, proc_asig, channel)
        output_path = os.path.join(args.img_dir,
                                   args.img_name.replace('_channel0', f'_channel{channel}'))
        save_plot(output_path)



"""
Plot the critical points in the optical flow vector field.
"""

import argparse
import numpy as np
from copy import copy
import matplotlib.pyplot as plt
from utils.io_utils import load_neo, save_plot
from utils.neo_utils import analogsignal_to_imagesequence

CLI = argparse.ArgumentParser()
CLI.add_argument("--data", nargs='?', type=str, required=True,
                 help="path to input data in neo format")
CLI.add_argument("--output", nargs='?', type=str, required=True,
                 help="path of output file")
CLI.add_argument("--skip_step", nargs='?', type=int, default=3,
                 help="skipping every x vector for the plot")
CLI.add_argument("--frame_id", nargs='?', type=int, default=0,
                 help="number of the frame to plot")

def plot_frame(frame, ax=None, skip_step=3):
    dim_y, dim_x = frame.shape

    if ax is None:
        fig, ax = plt.subplots()

    ax.quiver(np.arange(dim_x)[::skip_step],
              np.arange(dim_y)[::skip_step],
              np.real(frame[::skip_step,::skip_step]),
              np.imag(frame[::skip_step,::skip_step]))

    Y, X = np.meshgrid(np.arange(dim_y), np.arange(dim_x), indexing='ij')
    ZR = np.real(frame)
    ZI = np.imag(frame)
    contourR = ax.contour(X, Y, ZR, levels=[0], colors='b', label='x = 0')
    contourI = ax.contour(X, Y, ZI, levels=[0], colors='g', label='y = 0')
    return ax


if __name__ == '__main__':
    args, unknown = CLI.parse_known_args()
    block = load_neo(args.data)

    asigs = block.filter(name='optical_flow', objects="AnalogSignal")

    if asigs:
        asig = asigs[0]
    else:
        raise ValueError("Input does not contain a signal with name " \
                       + "'optical_flow'!")
   
    imgseq = analogsignal_to_imagesequence(asig)

    crit_point_evt = [evt for evt in block.segments[0].events
                      if evt.name == "critical_points"]
    if crit_point_evt:
        crit_point_evt = crit_point_evt[0]
    else:
        raise ValueError("Input does not contain an event with name " \
                       + "'critical_points'!")

    fig, ax = plt.subplots()
    
    ax = plot_frame(imgseq.as_array()[args.frame_id], 
                    skip_step=args.skip_step, 
                    ax=ax)

    start_id = np.argmax(crit_point_evt.times >= asig.times[args.frame_id])
    stop_id = np.argmax(crit_point_evt.times >= asig.times[args.frame_id+1])

    ax.scatter(crit_point_evt.array_annotations['x'][start_id:stop_id],
               crit_point_evt.array_annotations['y'][start_id:stop_id],
               color='r')

    ax.set_title(f"{asig.times[args.frame_id]:.2f}")
    ax.set_xticklabels([])
    ax.set_yticklabels([])

    save_plot(args.output)



"""
Create frames showing the input data on the spatial grid for each time step.
"""

import os
import sys
import argparse
import matplotlib.pyplot as plt
import numpy as np
from utils.io_utils import load_neo, save_plot
from utils.neo_utils import analogsignal_to_imagesequence
from utils.parse import none_or_str, none_or_float


CLI = argparse.ArgumentParser()
CLI.add_argument("--data", nargs='?', type=str,
                 help="path to input data in neo format")
CLI.add_argument("--frame_folder", nargs='?', type=str,
                 help="")
CLI.add_argument("--frame_name", nargs='?', type=str,
                 help="")
CLI.add_argument("--frame_format", nargs='?', type=str,
                 help="")
CLI.add_argument("--frame_rate", nargs='?', type=none_or_float,
                 help="")
CLI.add_argument("--colormap", nargs='?', type=str,
                 help="")
CLI.add_argument("--plot_event", nargs='?', type=none_or_str, default=None,
                 help="")
CLI.add_argument("--marker_color", nargs='?', type=str, default='k',
                 help="")


def get_events(events, frame_times, event_name='transitions'):
    trans_events = [ev for ev in events if ev.name == event_name]
    if len(trans_events):
        event = trans_events[0]
        ups = np.array([(t,
                         event.array_annotations['x_coords'][i],
                         event.array_annotations['y_coords'][i])
                         for i, t in enumerate(event)],
                         # if event.labels[i] == 'UP'],
                       dtype=[('time', 'float'),
                              ('x_coords', 'int'),
                              ('y_coords', 'int')])
        ups = np.sort(ups, order=['time', 'x_coords', 'y_coords'])

        up_coords = []
        for frame_count, frame_time in enumerate(frame_times):
            # select indexes of up events during this frame
            idx = range(np.argmax(np.bitwise_not(ups['time'] < frame_time)),
                        np.argmax(ups['time'] > frame_time))
            frame_ups = np.array([(x,y) for x, y in zip(ups['x_coords'][idx],
                                                        ups['y_coords'][idx])])
            up_coords.append(frame_ups)
        return up_coords
    else:
        print(f"No {event_name} events found!")
        return None


def get_opticalflow(block, imgseq_name="optical_flow"):
    asigs = block.filter(name=imgseq_name, objects="AnalogSignal")
    if len(asigs):
        # Normalize?
        return analogsignal_to_imagesequence(asigs[0])
    else:
        return None

def stretch_to_framerate(t_start, t_stop, num_frames, frame_rate=None):
    if frame_rate is None:
        return np.arange(num_frames, dtype=int)
    else:
        new_num_frames = (t_stop.rescale('s').magnitude
                        - t_start.rescale('s').magnitude) \
                        * frame_rate
        return np.linspace(0, num_frames-1, int(new_num_frames), dtype=int)

def plot_frame(frame, cmap=plt.cm.gray, vmin=None, vmax=None):
    fig, ax = plt.subplots()
    img = ax.imshow(frame, interpolation='nearest',
                    cmap=cmap, vmin=vmin, vmax=vmax, origin='lower')
    plt.colorbar(img, pad=0, ax=ax)

    ax.axis('image')
    ax.set_xticks([])
    ax.set_yticks([])
    # ax.set_xlim((0, dim_x))
    # ax.set_ylim((dim_y, 0))
    return ax

def plot_transitions(up_coords, markersize=1, markercolor='k', ax=None):
    if up_coords.size:
        if ax is None:
            ax = plt.gca()

        ax.plot(up_coords[:,0], up_coords[:,1],
                marker='D', color=markercolor, markersize=markersize,
                linestyle='None', alpha=0.6)
        # if len(pixels[0]) > 0.005*pixel_num:
        #     slope, intercept, _, _, stderr = scipy.stats.linregress(pixels[1], pixels[0])
        #     if stderr < 0.18:
        #         ax.plot(x, [intercept + slope*xi for xi in x], color='r')
    return ax

def plot_vectorfield(frame, skip_step=3, ax=None):
    # Every <skip_step> point in each direction.
    if ax is None:
        ax = plt.gca()
    dim_y, dim_x = frame.shape
    y_idx, x_idx = np.meshgrid(np.arange(dim_y), np.arange(dim_x), indexing='ij')
    xi = x_idx[::skip_step,::skip_step].flatten()
    yi = y_idx[::skip_step,::skip_step].flatten()
    ax.quiver(xi, yi, np.real(frame[yi,xi]), np.imag(frame[yi,xi]))
    return ax

if __name__ == '__main__':
    args, unknown = CLI.parse_known_args()

    blk = load_neo(args.data)
    asig = blk.segments[0].analogsignals[0]
    imgseq = analogsignal_to_imagesequence(asig)
    
    optical_flow = get_opticalflow(blk)

    # get data
    times = blk.segments[0].analogsignals[0].times  # to be replaced
    t_start = blk.segments[0].analogsignals[0].t_start  # to be replaced
    t_stop = blk.segments[0].analogsignals[0].t_stop  # to be replaced
    dim_t, dim_y, dim_x = imgseq.shape

    if args.plot_event is not None:
        up_coords = get_events(blk.segments[0].events,
                               frame_times=times,
                               event_name=args.plot_event)

    # prepare plotting
    frame_idx = stretch_to_framerate(t_start=t_start,
                                     t_stop=t_stop,
                                     num_frames=dim_t,
                                     frame_rate=args.frame_rate)
    if args.colormap == 'gray':
        cmap = plt.cm.gray
    else:
        # 'gray', 'viridis' (sequential), 'coolwarm' (diverging), 'twilight' (cyclic)
        cmap = plt.get_cmap(args.colormap)

    frames = imgseq.as_array()
    vmin = np.nanmin(frames)
    vmax = np.nanmax(frames)
    markersize = 50 / max([dim_x, dim_y])
    skip_step = int(min([dim_x, dim_y]) / 50) + 1

    # plot frames
    for i, frame_num in enumerate(frame_idx):
        ax = plot_frame(frames[frame_num], cmap=cmap, vmin=vmin, vmax=vmax)

        if optical_flow is not None:
            plot_vectorfield(optical_flow[frame_num].as_array(), 
                             skip_step=skip_step)
        if args.plot_event is not None and up_coords is not None:
            plot_transitions(up_coords[frame_num], markersize=markersize,
                             markercolor=args.marker_color)
        ax.set_ylabel('pixel size: {:.2f} mm'.format(imgseq.spatial_scale.rescale('mm').magnitude))
        ax.set_xlabel('{:.3f} s'.format(times[frame_num].rescale('s').magnitude))

        save_plot(os.path.join(args.frame_folder,
                               args.frame_name + '_{}.{}'.format(str(i).zfill(5),
                               args.frame_format)),
                  transparent=True)
        plt.close()
 



"""
Calculate the wave propagation velocity for each wave.
"""

from pathlib import Path
import numpy as np
import matplotlib.pyplot as plt
import argparse
import scipy
import pandas as pd
from utils.io_utils import load_neo, save_plot
from utils.parse import none_or_str

CLI = argparse.ArgumentParser()
CLI.add_argument("--data", nargs='?', type=Path, required=True,
                 help="path to input data in neo format")
CLI.add_argument("--output", nargs='?', type=Path, required=True,
                 help="path of output file")
CLI.add_argument("--output_img", nargs='?', type=none_or_str, default=None,
                 help="path of output image file")
CLI.add_argument("--event_name", "--EVENT_NAME", nargs='?', type=str, default='wavefronts',
                 help="name of neo.Event to analyze (must contain waves)")

def center_points(x, y):
    return x - np.mean(x), y - np.mean(y)

def linregress(times, locations):
    times, locations = center_points(times, locations)
    slope, offset, _, _, stderr = scipy.stats.linregress(times, locations)
    return slope, stderr, offset

def calc_planar_velocities(evts):
    spatial_scale = evts.annotations['spatial_scale']
    v_unit = (spatial_scale.units/evts.times.units).dimensionality.string

    wave_ids = np.unique(evts.labels)

    velocities = np.zeros((len(wave_ids), 2)) * np.nan

    ncols = int(np.round(np.sqrt(len(wave_ids)+1)))
    nrows = int(np.ceil((len(wave_ids)+1)/ncols))
    fig, ax = plt.subplots(nrows=nrows, ncols=ncols,
                           figsize=(3*nrows, 3*ncols))

    # loop over waves
    for i, wave_i in enumerate(wave_ids):
        # Fit wave displacement
        idx = np.where(evts.labels == wave_i)[0]
        times = evts.times[idx].magnitude
        if (times == times[0]).all():
            continue
        x_times, x_locations = center_points(times,
                                        evts.array_annotations['x_coords'][idx]
                                        * spatial_scale.magnitude)
        y_times, y_locations = center_points(times,
                                        evts.array_annotations['y_coords'][idx]
                                        * spatial_scale.magnitude)
        vx, vx_err, dx = linregress(x_times, x_locations)
        vy, vy_err, dy = linregress(y_times, y_locations)
        v = np.sqrt(vx**2 + vy**2)
        v_err = 1/v * np.sqrt((vx*vx_err)**2 + (vy+vy_err)**2)
        velocities[i] = np.array([v, v_err])

        # Plot fit
        row = int(i/ncols)
        if ncols == 1:
            cax = ax[row]
            col = 0
        else:
            col = i % ncols
            cax = ax[row][col]
        cax.plot(x_times, x_locations,
                color='b', label='x coords', linestyle='', marker='.', alpha=0.5)
        cax.plot(x_times, [vx*t + dx for t in x_times], color='b')
        cax.plot(y_times, y_locations,
                color='r', label='y coords', linestyle='', marker='.', alpha=0.5)
        cax.plot(y_times, [vy*t + dy for t in y_times], color='r')
        if not col:
            cax.set_ylabel('x/y position [{}]'\
                           .format(spatial_scale.dimensionality.string))
        if row == nrows-1:
            cax.set_xlabel('time [{}]'\
                           .format(evts.times[idx].dimensionality.string))
        cax.set_title('wave {}'.format(wave_i))

    # plot total velocities
    if ncols == 1:
        cax = ax[-1]
    else:
        cax = ax[-1][-1]
        for i in range(len(wave_ids), nrows*ncols-1):
            row = int(i/ncols)
            col = i % ncols
            ax[row][col].set_axis_off()

    cax.errorbar(wave_ids, velocities[:,0], yerr=velocities[:,1],
                        linestyle='', marker='+')
    cax.set_xlabel(f'{evts.name} id')
    cax.set_title('velocities [{}]'.format(v_unit))

    # transform to DataFrame
    df = pd.DataFrame(velocities,
                      columns=['velocity_planar', 'velocity_planar_std'])
    df['velocity_unit'] = v_unit
    return df


if __name__ == '__main__':
    args, unknown = CLI.parse_known_args()

    block = load_neo(args.data)

    evts = block.filter(name=args.event_name, objects="Event")[0]
    evts = evts[evts.labels.astype('str') != '-1']

    velocities_df = calc_planar_velocities(evts)
    velocities_df[f'{args.event_name}_id'] = np.unique(evts.labels)

    if args.output_img is not None:
        save_plot(args.output_img)

    velocities_df.to_csv(args.output)



"""
Check whether the input data representation adheres to the stage's requirements.

Additionally prints a short summary of the data attributes.
"""

import numpy as np
import argparse
import quantities as pq
import warnings
import re
from utils.io_utils import load_neo
from utils.parse import none_or_str

CLI = argparse.ArgumentParser()
CLI.add_argument("--data", nargs='?', type=str, required=True,
                 help="path to input data in neo format")
CLI.add_argument("--event_name", "--EVENT_NAME", nargs='?', type=none_or_str, default=None,
                 help="name of neo.Event to analyze (must contain waves)")
CLI.add_argument("--measures", "--MEASURES", nargs='+', type=none_or_str, default=None,
                 help="list of measure names to apply")

if __name__ == '__main__':
    args, unknown = CLI.parse_known_args()

    block = load_neo(args.data)

    if len(block.segments) > 1:
        print("More than one Segment found; all except the first one " \
            + "will be ignored.")

    evts = block.filter(name='wavefronts', objects="Event")
    if not len(evts):
        raise ValueError("No 'wavefronts' events found!")

    evt = evts[0]
    evt = evt[evt.labels != '-1']
    num_waves = len(np.unique(evt.labels))

    if num_waves:
        print(f'{num_waves} wavefronts found.')
    else:
        raise ValueError("There are no waves detected!")

    evt.array_annotations['x_coords']
    evt.array_annotations['y_coords']
    evt.annotations['spatial_scale']

    optical_flow = block.filter(name='optical_flow', objects="AnalogSignal")
    if not len(optical_flow):
        warnings.warn('No Optical-Flow signal available!')

    evts = block.filter(name='wavemodes', objects="Event")
    if len(evts):
        print(f'{len(np.unique(evts[0].labels))} wavemodes found')
    else:
        warnings.warn("No 'wavemodes' events found!")



import contextlib
import inspect
import logging
import os
import re
import shutil
import sys
from pathlib import Path
from pprint import pformat
from string import ascii_lowercase

sys.path.append(str(Path(inspect.getfile(lambda: None))))
from pipeline.utils.snakefile import (
    get_config,
    get_setting,
    load_config_file,
    locate_str_in_list,
    read_stage_output,
    set_setting,
    update_configfile,
)

log = logging.getLogger()
logging.basicConfig(level=logging.INFO)


def create_new_configfile(profile, stage=None, stage_number=None, parent=None):
    if stage is None:
        if stage_number is None:
            raise KeyError("You must either specify `stage` or `stage_number`!")
        else:

            stage = get_setting("stages")[stage_number]

    config_path = Path(get_setting("config_path"))

    parent = f"_{parent}" if parent else ""
    config_dir = config_path / stage / "configs"
    template_config_path = config_dir / f"config{parent}.yaml"
    new_config_path = config_dir / f"config_{profile}.yaml"

    if new_config_path.exists():
        log.debug(f"config file `{new_config_path}` already exists. Skip.")
        return None

    if not template_config_path.exists():
        log.debug(f"parent config file `{template_config_path}` doesn't exist.")
        potential_parents = [
            f.name
            for f in config_dir.iterdir()
            if f.name.startswith(f"config{parent}")
        ]
        if potential_parents:
            log.debug(f"Using `{potential_parents[0]}` instead.")
            template_config_path = config_dir / potential_parents[0]
        else:
            raise FileNotFoundError(
                "No parent config file fitting the name "
                f"`{parent.strip('_')}` found!"
            )

    shutil.copy(template_config_path, new_config_path)
    update_configfile(new_config_path, {"PROFILE": profile})
    return None


def input_profile(profile=None):
    if not is_profile_name_valid(profile) and profile is not None:
        log.info(f"profile name {profile} is not valid!")
        profile = None
    if profile is None:
        while not is_profile_name_valid(profile):
            profile = input("profile name: ")
            if not is_profile_name_valid(profile):
                log.info(
                    f"{profile} is not a valid profile name. "
                    "Please use only letters, "
                    "'_' (for sub-profile structuring), "
                    "or '|' (for variant separation)."
                )
    return profile


def input_stage(stage=None):
    stages = get_setting("stages")

    while stage not in stages.keys() and stage not in stages.values():
        stage = input(
            "Which stage should be executed?\n    "
            + "\n    ".join(f"{k} {v}" for k, v in stages.items())
            + "\nSelect the stage index or the stage name: "
        )
    if stage in stages.keys():
        stage = stages[stage]
    return stage


def input_block(stage=None, block=None):
    if stage is None:
        raise ValueError("Stage must be specified!")

    block_dir = Path(get_setting("pipeline_path")) / stage / "scripts"
    available_blocks = get_available_blocks(block_dir)

    while block not in available_blocks:

        if block:
            print(f"Block '{block}' is not found in {stage}!")
        else:
            print("Which block should be executed?")

        print(f"Available blocks are: {', '.join(available_blocks)}")
        block = input("Select block: ")

    return block


def print_settings(*args, **kwargs):
    print(pformat(get_setting()))
    return None


def get_available_blocks(block_dir):
    available_scripts = [
        script for script in block_dir.iterdir() if os.path.isfile(script)
    ]
    available_blocks = [
        s.stem
        for s in available_scripts
        if not (s.stem).startswith("_")
        and s.suffix == ".py"
        and "template" not in s.stem
    ]
    return available_blocks


def is_profile_name_valid(profile: str) -> bool:
    if isinstance(profile, str):
        profile = profile.strip("'\"")
        pattern = re.compile("[\w\d\|]+")
        return bool(pattern.fullmatch(profile))
    else:
        return False


def get_profile(profile=None, parent_profile=None):
    # set initial profile name
    profile = input_profile(profile=profile)

    # set full profile name
    if parent_profile is None:
        parent_profile = input("Specify the parent profile name [optional]:")

    if (
        is_profile_name_valid(parent_profile)
        and "|" not in parent_profile
        and not profile.startswith(parent_profile)
    ):
        profile = f"{parent_profile}_{profile}"
    elif parent_profile:
        log.info(f"{parent_profile} is not a valid profile parent name. Ignore.")
        parent_profile = None
    return profile, parent_profile


def setup_entry_stage(
    profile, parent_profile=None, data_path=None, loading_script_name=None
):
    """
    Populate an existing config file for stage 01 with a data path and create
    a corresponding loading script from a template.
    """
    config_path = Path(get_setting("config_path"))

    stage01_update_dict = {}

    # set data path
    if data_path is None:
        data_path = (
            input(
                "Path to dataset [optional, "
                "can be set later in the stage01 config file]:"
            )
            or None
        )
    if data_path is not None:
        stage01_update_dict.update({"DATA_SETS": {profile: data_path}})

    # set loading script path
    if loading_script_name is None:
        base_profile = parent_profile if parent_profile else profile

        loading_script_name = (
            input(
                "Loading script name (to be created) "
                "in <config_path>/stage01_data_entry/scripts/ "
                "[default: enter_<profile>.py]:"
            )
            or f"enter_{base_profile}.py"
        )
        if not Path(loading_script_name).suffix == ".py":
            loading_script_name = f"{str(Path(loading_script_name).stem)}.py"
        stage01_update_dict.update({"CURATION_SCRIPT": loading_script_name})

    stages = get_setting("stages")
    loading_script_path = (
        config_path / stages["1"] / "scripts" / f"{loading_script_name}"
    )
    if loading_script_path.exists():
        log.info(f"`{loading_script_name}` already exists. Skip.")
    else:
        shutil.copy(
            config_path / stages["1"] / "scripts" / "enter_data_template.py",
            loading_script_path,
        )

    # update stage 01 config
    stage01_config_path = (
        config_path / stages["1"] / "configs" / f"config_{profile}.yaml"
    )
    update_configfile(stage01_config_path, stage01_update_dict)
    return None


def get_initial_available_stages():
    pipeline_path = Path(get_setting("pipeline_path"))
    stages = [
        x.name
        for x in pipeline_path.iterdir()
        if x.is_dir() and str(x.name).startswith("stage") and "template" not in x.name
    ]
    stages.sort()
    stage_numbers = [int(re.findall(r"\d+", stage)[0]) for stage in stages]
    stage_keys = []
    for i, v in enumerate(stage_numbers):
        totalcount = stage_numbers.count(v)
        count = stage_numbers[:i].count(v)
        stage_keys.append(
            str(v) + ascii_lowercase[count] if totalcount > 1 and count > 0 else str(v)
        )
    return {k: v for k, v in zip(stage_keys, stages)}


@contextlib.contextmanager
def working_directory(path):
    """Changes working directory and returns to previous on exit."""
    prev_cwd = Path.cwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(prev_cwd)
    return None



"""
Determine the threshold between Up and Down states for each channel
by fitting the respective amplitude distributions.
"""

import numpy as np
import argparse
import os
import scipy as sc
import warnings
from scipy.optimize import OptimizeWarning
import matplotlib.pyplot as plt
from utils.io_utils import load_neo, save_plot
from utils.parse import none_or_int

CLI = argparse.ArgumentParser()
CLI.add_argument("--data", nargs='?', type=str, required=True,
                 help="path to input data in neo format")
CLI.add_argument("--output", nargs='?', type=str, required=True,
                 help="path of output thresholds (numpy array)")
CLI.add_argument("--img_dir", nargs='?', type=str, default=None,
                 help="path of output figure directory")
CLI.add_argument("--img_name", nargs='?', type=str, default=None,
                 help="example name of output figure for channel 0")
CLI.add_argument("--fit_function", nargs='?', type=str, default='Gaussian',
                 help="function to fit the amplitude distribution")
CLI.add_argument("--sigma_factor", nargs='?', type=float, default=3,
                 help="sigma_factor x standard deviation = threshold")
CLI.add_argument("--bin_num", nargs='?', type=int, default=100,
                 help='number of bins for the amplitude histogram')
CLI.add_argument("--plot_channels", nargs='+', type=none_or_int,
                 default=None, help="list of channels to plot")

warnings.simplefilter("error", OptimizeWarning)

def gaussian(x, mu=0, sig=1):
    return 1. / (sig * np.sqrt(2 * np.pi)) \
         * np.exp(-np.power(x - mu, 2.) / (2 * np.power(sig, 2.)))

def double_gaussian(x, ratio=.5, mu1=0, sig1=1, mu2=0, sig2=1):
    return ratio * gaussian(x=x, mu=mu1, sig=sig1) \
         + (1-ratio) * gaussian(x=x, mu=mu2, sig=sig2)

def double_gaussian_fit(params, x, y):
    ratio, mu1, sig1, mu2, sig2 = params
    fit = double_gaussian(x, ratio=ratio, mu1=mu1, sig1=sig1, mu2=mu2, sig2=sig2)
    return fit - y


def fit_amplitude_distribution(signal, sigma_factor, fit_function,
                               bins, plot_channel):
    # signal amplitude distribution
    signal = signal[np.isfinite(signal)]
    hist, edges = np.histogram(signal, bins=bins, density=True)
    xvalues = edges[:-1] + np.diff(edges)[0] / 2.

    if fit_function == 'HalfGaussian':
        # First fit -> determine peak location m0
        try:
            (m0, _), _ = sc.optimize.curve_fit(gaussian, xvalues, hist, p0=(0, 1))
        except OptimizeWarning:
            print('Could not perform first fit. Using Median to determine central downstate signal amplitude.')
            m0 = np.median(signal)

        # shifting to 0
        xvalues -= m0

        # Mirror left peak side for 2nd fit
        signal_leftpeak = signal[signal - m0 <= 0] - m0
        mirror_signal = np.append(signal_leftpeak, -1 * signal_leftpeak)
        peakhist, edges = np.histogram(mirror_signal, bins=bins, density=True)
        xvalues2 = edges[:-1] + np.diff(edges)[0] / 2.

        # Second fit -> determine spread s0
        try:
            (_, s0), _ = sc.optimize.curve_fit(fit_function, xvalues2, peakhist, p0=(0, np.std(peakhist)))
        except OptimizeWarning:
            print('Could not perform second fit. Using std to determine spread of downstate signal amplitudes.')
            s0 = np.std(peakhist)
        except RuntimeError:
            print('Could not perform second fit. Using std to determine spread of downstate signal amplitudes.')
            s0 = np.std(peakhist)

        if plot_channel:
            fig, ax = plt.subplots(figsize=(7, 7))
            ax.bar(xvalues, hist, width=np.diff(xvalues)[0], color='r')
            left_right_ratio = len(signal_leftpeak) * 2. / len(signal)
            ax.plot(xvalues, [left_right_ratio * fit_function(x, 0, s0) for x in xvalues], c='k')
            ax.set_xlabel('signal')
            ax.set_ylabel('sample density')
            ax.set_title('Amplitude distribution (channel {})'.format(plot_channel))
            ax.axvline(sigma_factor * s0, color='k', ls='--'),
            ax.text(1.1 * sigma_factor * s0, 0.9 * ax[1].get_ylim()[0],
                       r'UD threshold ({}$\sigma$)'.format(sigma_factor), color='k')
        return m0 + sigma_factor * s0

    elif fit_function == 'DoubleGaussian':
        # ratio, mu1, sig1, mu2, sig2
        inital_params = np.array([.5, xvalues[0], .1, xvalues[-1], .1])
        mu_bound = (xvalues[0], xvalues[-1])
        sig_bound = (0, xvalues[-1]-xvalues[0])
        bounds = ([0, mu_bound[0], sig_bound[0], mu_bound[0], sig_bound[0]],
                  [1, mu_bound[1], sig_bound[1], mu_bound[1], sig_bound[1]])

        optimize_result = sc.optimize.least_squares(double_gaussian_fit,
                                                    inital_params,
                                                    bounds=bounds,
                                                    kwargs={'x':xvalues,
                                                            'y':hist})
        ratio, mu1, sig1, mu2, sig2 = optimize_result.x

        x = np.linspace(mu1, mu2, 100)
        min_idx = sc.signal.argrelmin(double_gaussian(x, ratio=ratio, mu1=mu1,
                                                      sig1=sig1, mu2=mu2, sig2=sig2))
        if len(min_idx[0]):
            threshold = x[min_idx[0][0]]
        else:
            threshold = mu1 + sigma_factor*sig1

        if plot_channel:
            fig, ax = plt.subplots(figsize=(7, 7))
            ax.bar(xvalues, hist, width=np.diff(xvalues)[0], color='r')
            ax.plot(xvalues, double_gaussian(xvalues, ratio=ratio, mu1=mu1, sig1=sig1, mu2=mu2, sig2=sig2))
            ax.axvline(threshold, linestyle=':', color='k')
        return threshold
    else:
        raise NotImplementedError('The fitting function {} is not yet implementd!'\
                                  .format(fit_function))

if __name__ == '__main__':
    args, unknown = CLI.parse_known_args()
    
    asig = load_neo(args.data, 'analogsignal')
    signal = asig.as_array()
    dim_t, dim_channels = signal.shape
    non_nan_channels = [i for i in range(dim_channels) if np.isfinite(signal[:,i]).all()]

    thresholds = np.empty(dim_channels)
    thresholds.fill(np.nan)

    for channel in non_nan_channels:
        if channel in args.plot_channels:
            plot_channel = channel
        else:
            plot_channel = False
        thresholds[channel] = fit_amplitude_distribution(signal[:,channel],
                                                         args.sigma_factor,
                                                         args.fit_function,
                                                         args.bin_num,
                                                         plot_channel)
        if plot_channel:
            output_path = os.path.join(args.img_dir,
                                       args.img_name.replace('_channel0', f'_channel{channel}'))
            save_plot(output_path)

    np.save(args.output, thresholds)



import os
from pathlib import Path
from ruamel.yaml import YAML
from copy import copy
from types import SimpleNamespace
from snakemake.logging import logger
yaml = YAML()
yaml.indent(mapping=4, sequence=4, offset=4)
# Todo: re-evaluate if this is the most elegant solution to hardcode
#       the settings path here
SETTINGS_PATH = Path('~/.cobrawap/config').expanduser()

def get_setting(key: str = None):
    with open(SETTINGS_PATH, 'r') as f:
        try:
            settings = yaml.load(f)
            if not settings:
                settings = {}
        except Exception as e:
            logger.warning(e)
            raise FileNotFoundError("The settings file either doesn't exist "\
                                    "or is not in proper JSON format! "\
                                    "Try to run `cobrawap init` to set up the "\
                                    "settings file.")
    if key is None:
        return settings
    elif key not in settings.keys():
        raise ValueError(f"{key} not found in settings file!")
    else:
        return settings[key]


def set_setting(setting: dict) -> None:
    if type(setting) != dict:
        raise TypeError('Function expects a dictionary!')
    
    if SETTINGS_PATH.exists():
        with open(SETTINGS_PATH, 'r') as f:
            settings = yaml.load(f)
        if not settings:
            settings = {}
    else:
        logger.warning("Creating a new cobrawap settings file at "\
                      f"`{SETTINGS_PATH.resolve()}`.")
        if not SETTINGS_PATH.parent.exists():
            SETTINGS_PATH.parent.mkdir(parents=True)   
        SETTINGS_PATH.touch(exist_ok=True)
        settings = {}
    
    key_overlap = [k for k in setting.keys() if k in settings.keys()]
    if key_overlap:
        overwrite = (input(f"There are already settings for {key_overlap}! "\
                            "Overwrite? [y/N]").lower() == 'y'
                    or False)
        if not overwrite:
            _ = [setting.pop(k, None) for k in key_overlap]

    settings.update(setting)
    with open(SETTINGS_PATH, 'w') as f:
         yaml.dump(settings, f)
    return None


def safe_open_w(path):
    os.makedirs(os.path.dirname(path), exist_ok=True)
    return open(path, 'w')


def read_stage_output(stage, config_dir, config_name,
                      output_namespace="STAGE_OUTPUT"):
    config_path = Path(config_dir) / stage
    if (config_path / 'configs').is_dir():
        config_path = config_path / 'configs'

    with open(config_path / config_name, 'r') as f:
        config_dict = yaml.load(f)
    if config_dict is None:
        logger.warning('config file '
                      f'{os.path.join(config_dir, stage, config_name)} '
                       'can not be loaded! Skipping reading stage output.')
        return None
    elif output_namespace in config_dict.keys():
        stage_output = config_dict[output_namespace]
        if not Path(stage_output).suffix:
            neo_format = config_dict['NEO_FORMAT']
            stage_output = f'{stage_output}.{neo_format}'
        return stage_output
    else:
        logger.error(f"config file of stage {stage} "
                     f"does not define {output_namespace}!")


def load_config_file(config_path):
    with open(config_path, 'r') as f:
        config_dict = yaml.load(f)
    if not config_dict:
        logger.error(f"config file is empty: {config_path}")
        raise FileNotFoundError
    return config_dict


def get_parent_config_name(config_name):
    """
    splitting last '_<something>' element
    keeping a trailing '|<anything>'
    """
    name, ext = os.path.splitext(config_name)
    main, *variant = name.split('|')
    variant = variant[0] if variant else ''

    parent = "_".join(main.split('_')[:-1])
    if not parent:
        return False
    elif variant:
        parent = '|'.join([parent, variant])

    return parent + ext


def get_config(config_dir, config_name, get_path_instead=False):
    """
    # search order:
    config_some_profile_name|variant.yaml
    config_some_profile|variant.yaml
    config_some|variant.yaml
    config|variant.yaml
    config_some_profile_name.yaml
    config_some_profile.yaml
    config_some.yaml
    config.yaml
    """
    config_dict = {}
    config_dir = Path(config_dir)
    if (config_dir / 'configs').is_dir():
        config_dir = config_dir / 'configs'

    try_config_name = copy(config_name)

    keep_variant = True
    while not config_dict:
        try:
            config_path = config_dir / try_config_name
            config_dict = load_config_file(config_path)
        except FileNotFoundError:
            prev_try_config_name = try_config_name

            parent_config_name = get_parent_config_name(try_config_name)
            if parent_config_name:
                try_config_name = parent_config_name
            else:
                if keep_variant and '|' in try_config_name:
                    name, ext = os.path.splitext(config_name)
                    try_config_name = name.split('|')[0] + ext
                    keep_variant = False
                else:
                    logger.error("No corresponding config file found!")
                    return {}

            logger.info(f"'{prev_try_config_name}' not found, "
                        f"trying '{try_config_name}'.")
            
    logger.info(f"Using '{try_config_name}.")
    if get_path_instead:
        return config_path
    else:
        return config_dict


def create_temp_configs(stages, configs_dir, config_name, output_dir,
                        temp_name='temp_config.yaml'):
    for i, stage in enumerate(stages):
        config_dict = get_config(os.path.join(configs_dir, stage), config_name)
        new_config_path = os.path.join(output_dir, stage, temp_name)
        with safe_open_w(new_config_path) as f:
            yaml.dump(config_dict, f)
    return None


def update_configfile(config_path, update_dict):
    # Careful! This function overwrites the config file.
    # Comments in the file are lost.
    with open(config_path, 'r') as f:
        try:
            config_dict = yaml.load(f)
        except Exception as e:
            logger.info(e)
            config_dict = None
    if config_dict is None:
        logger.warning(f'config file {config_path} can not be loaded! '
                       'Skipping updating config.')
        return None
    config_dict.update(**update_dict)
    with safe_open_w(config_path) as f:
        yaml.dump(config_dict, f)
    return None


def set_stage_inputs(stages, output_dir, config_file='temp_config.yaml',
                     input_namespace="STAGE_INPUT"):
    update_dict = {}
    update_configfile(config_path=os.path.join(output_dir, stages[0], config_file),
                      update_dict={input_namespace:None})

    for i, stage in enumerate(stages[:-1]):
        output_name = read_stage_output(stage,
                                        config_dir=output_dir,
                                        config_name=config_file)
        if output_name is None:
            logger.warning(f'Could not read stage output for {stage}! '
                           'Skipping setting input for subsequent stage.')
        else:
            update_dict[input_namespace] = os.path.join(output_dir, stage, output_name)
            update_configfile(config_path=os.path.join(output_dir, stages[i+1], config_file),
                              update_dict=update_dict)
    return None


def set_global_configs(stages, output_dir, config_dict,
                       config_file='temp_config.yaml'):
    for stage in stages:
        update_configfile(config_path=os.path.join(output_dir, stage, config_file),
                          update_dict=config_dict)
    return None


def dict_to_cla(arg_dict):
    for key, value in arg_dict.items():
        if type(value) == list:
            arg_dict[key] = ' '.join(str(v) for v in value)

    cla_str = lambda k,v: f'--{k} "{v}"' if in_quotes(v) else f'--{k} {v}'
    arg_strings = [cla_str(key, value) for key, value in arg_dict.items()]
    return ' '.join(arg_strings)


def in_quotes(object):
    char = ['\s', '|']  # when these characters are in an argument str
                        # they need to be put in quotes to be passed as cla
    return any([c in str(object) for c in char])


def params(*args, config=None, **kwargs):
    '''
    Creates a parameter dictionary that is returned as a string as
    command line argument as `--key value` for each key-value pair in the dict.
    
    args:
        - if args[0] is a dict, it is added to the parameter dict
        - if args are strings and config is a dict, args (and args.upper()) 
          are interpreted as keys of config dict and the corresponding elements 
          are added to the parameter dict
    config:
        dict from which elements (keys given by args) are selected
    kwargs:
        kwargs are added as key-value pairs are to the parameter dict

    Additionally, all wildcards and named outputs of the current rule are added 
    as key-value pairs to the parameter dict.
    '''

    param_dict = {}  # use default dict?

    if len(args) and type(args[0]) == dict:
        param_dict = args[0]

    if type(config) == SimpleNamespace:
        config = vars(config)

    if type(config) == dict:
        for arg in args:
            if not type(arg) == str:
                continue
            if arg in config.keys():
                param_dict[arg] = config[arg]
            elif arg.upper() in config.keys():
                param_dict[arg] = config[arg.upper()]
            else:
                logger.info(f'Parameter "{arg}" not found in the config! ' 
                             'Set to None!')
                param_dict[arg] = None

    param_dict.update(dict(kwargs.items()))

    def add_output_and_wildcards_to_args(wildcards, output):
        for items in [wildcards, output]:
            item_dict = dict(items.items())
            if 'data' in item_dict.keys():
                logger.warning("wildcards or outputs named 'data' are being ignored!")
                del item_dict['data']

            duplicates = [key for key in item_dict.keys() if key in param_dict.keys()]
            
            for key in duplicates:
                if param_dict[key] != item_dict[key]:
                    logger.warning(f"The keyword {key} is used multiple times "
                                    "in the rule's params, wildcards, or output! \n"
                                   f"{key}: '{param_dict[key]}' is ignored "
                                   f"in favor of '{item_dict[key]}'")

            param_dict.update(item_dict)
        return dict_to_cla(param_dict) 

    return add_output_and_wildcards_to_args


def locate_str_in_list(str_list, string):
    if string in str_list:
        return [i for i, el in enumerate(str_list) if el == string][0]
    else:
        logger.error(f"Can't find rule '{string}'! Please check the spelling "
                      "and the config file.")



"""
Detrend the signal in each channel by order 0 (constant) or 1 (linear).
"""

import numpy as np
import scipy
import matplotlib.pyplot as plt
import argparse
import os
import warnings
from utils.io_utils import load_neo, write_neo, save_plot
from utils.parse import none_or_int

CLI = argparse.ArgumentParser()
CLI.add_argument("--data",    nargs='?', type=str, required=True,
                 help="path to input data in neo format")
CLI.add_argument("--output",  nargs='?', type=str, required=True,
                 help="path of output file")
CLI.add_argument("--order", nargs='?', type=int, default=1,
                 help="detrending order")
CLI.add_argument("--img_dir",  nargs='?', type=str, required=True,
                 help="path of output figure directory")
CLI.add_argument("--img_name", nargs='?', type=str,
                 default='processed_trace_channel0.png',
                 help='example filename for channel 0')
CLI.add_argument("--plot_channels", nargs='+', type=none_or_int, default=None,
                 help="list of channels to plot")

def detrend(asig, order):
    if (order != 0) and (order != 1):
        warnings.warn("Detrending order must be either 0 (constant) or 1 (linear)! Skip.")
        return asig

    dtrend = 'linear' if order else 'constant'
    detrended_signals = np.empty(asig.shape)
    detrended_signals.fill(np.nan)

    for channel in range(asig.shape[1]):
        channel_signal = asig.as_array()[:,channel]
        if np.isnan(channel_signal).any():
            continue
        detrended = scipy.signal.detrend(channel_signal, type=dtrend, axis=0)
        detrended_signals[:,channel] = detrended
    detrend_asig = asig.duplicate_with_new_data(detrended_signals)
    detrend_asig.array_annotate(**asig.array_annotations)
    return detrend_asig


def plot_detrend(asig, detrend_asig, channel):
    fig, ax = plt.subplots(nrows=2, sharex=True, figsize=(17,8))

    ax[0].plot(asig.times, asig.as_array()[:,channel], color='b', linewidth=1)
    ax[0].axhline(0, linestyle="--", color='k', linewidth=1)
    ax[0].set_ylabel('signal')

    ax[1].plot(asig.times, detrend_asig.as_array()[:,channel], color='g', linewidth=1)
    ax[1].axhline(0, linestyle="--", color='k', linewidth=1)
    ax[1].set_ylabel('detrended signal')
    ax[1].set_xlabel(f'time [{asig.times.dimensionality.string}]')
    return ax


if __name__ == '__main__':
    args, unknown = CLI.parse_known_args()

    block = load_neo(args.data)
    asig = block.segments[0].analogsignals[0]

    detrend_asig = detrend(asig, args.order)

    if args.plot_channels[0] is not None:
        for channel in args.plot_channels:
            plot_detrend(asig, detrend_asig, channel)
            output_path = os.path.join(args.img_dir,
                                       args.img_name.replace('_channel0', f'_channel{channel}'))
            save_plot(output_path)

    detrend_asig.description += "Detrended by order {} ({}). "\
                        .format(args.order, os.path.basename(__file__))
    block.segments[0].analogsignals[0] = detrend_asig

    write_neo(args.output, block)



"""
Estimate the logarithmic multi-unit activity (MUA) by averaging the power
in a given frequency range.
"""

import numpy as np
from elephant.spectral import welch_psd
from elephant.signal_processing import butter
from scipy.stats import zscore
import quantities as pq
import matplotlib.pyplot as plt
import seaborn as sns
import argparse
import os
from pathlib import Path
from utils.io_utils import load_neo, write_neo, save_plot
from utils.parse import none_or_float, none_or_int
from utils.neo_utils import time_slice

CLI = argparse.ArgumentParser()
CLI.add_argument("--data", nargs='?', type=Path, required=True,
                 help="path to input data in neo format")
CLI.add_argument("--output", nargs='?', type=Path, required=True,
                 help="path of output file")
CLI.add_argument("--img_dir", nargs='?', type=Path,
                 default=None, help="path of figure directory")
CLI.add_argument("--img_name", nargs='?', type=str,
                 default='minima_channel0.png',
                 help='example image filename for channel 0')
CLI.add_argument("--highpass_frequency", nargs='?', type=float, default=200,
                 help="lower bound of frequency band in Hz")
CLI.add_argument("--lowpass_frequency", nargs='?', type=float, default=1500,
                 help="upper bound of frequency band in Hz")
CLI.add_argument("--logMUA_rate", nargs='?', type=none_or_float, default=None,
                 help="rate of the signal after transformation")
CLI.add_argument("--psd_overlap", nargs='?', type=float, default=0.5,
                 help="overlap parameter for Welch's algorithm [0-1]")
CLI.add_argument("--fft_slice", nargs='?', type=none_or_float, default=None,
                 help="time window length used for power spectrum estimate, in s")
CLI.add_argument("--plot_tstart", nargs='?', type=none_or_float, default=0,
                 help="start time in seconds")
CLI.add_argument("--plot_tstop", nargs='?', type=none_or_float, default=10,
                 help="stop time in seconds")
CLI.add_argument("--plot_channels", nargs='+', type=none_or_int, default=None,
                 help="list of channels to plot")


def logMUA_estimation(asig, highpass_frequency, lowpass_frequency, logMUA_rate,
                      psd_overlap, fft_slice):
    time_steps, channel_num = asig.shape
    fs = asig.sampling_rate.rescale('Hz')
    if fft_slice is None:
        fft_slice = (1/highpass_frequency).rescale('s')
    elif fft_slice < 1/highpass_frequency:
        raise ValueError("Too few fft samples to estimate the frequency "\
                       + "content in the range [{} {}]."\
                         . format(highpass_frequency, lowpass_frequency))
    # logMUA_rate can only be an int fraction of the orginal sampling_rate
    if logMUA_rate is None:
        logMUA_rate = highpass_frequency
    if logMUA_rate > fs:
        raise ValueError("The requested logMUA rate can not be larger than "\
                       + "the inital sampling rate!")
    subsample_order = int(fs/logMUA_rate)
    eff_logMUA_rate = fs/subsample_order
    print("effective logMUA rate = {}".format(eff_logMUA_rate))

    if 1/eff_logMUA_rate > fft_slice:
        raise ValueError("The given logMUA_rate is too low to capture "\
                       + "the logMUA estimate of all the signal with "\
                       + "sample size {}. ".format(fft_slice)\
                       + "Either increase the logMUA_rate "\
                       + "or increase fft_slice.")

    subsample_times = np.arange(asig.t_start.rescale('s'),
                                asig.t_stop.rescale('s'),
                                1/eff_logMUA_rate) * asig.t_start.units

    non_nan_channels = [i for i in range(asig.shape[-1]) if np.isfinite(asig[:,i]).all()]
    asig_channels = asig[:, non_nan_channels]
    logMUA_signal = np.zeros((len(subsample_times), len(non_nan_channels)))

    for i, t in enumerate(subsample_times):
        if t < asig.t_start.rescale('s') + fft_slice/2:
            t_start = asig.t_start.rescale('s')
        elif t > asig.t_stop.rescale('s') - fft_slice/2:
            t_start = asig.t_stop.rescale('s') - fft_slice
        else:
            t_start = t - fft_slice/2

        t_stop = np.min([t_start + fft_slice,
                         asig.t_stop.rescale('s')]) *pq.s

        asig_slice = asig_channels.time_slice(t_start=t_start, t_stop=t_stop)

        freqs, psd = welch_psd(asig_slice,
                               frequency_resolution=highpass_frequency,
                               overlap=psd_overlap,
                               window='hann',
                               detrend='linear',
                               nfft=None)

        high_idx = (np.abs(freqs - lowpass_frequency)).argmin()
        if not i:
            print("logMUA signal estimated in frequency range "\
                + "{:.2f} - {:.2f}.".format(freqs[1], freqs[high_idx]))

        avg_power = np.mean(psd, axis=-1)
        # psd[:,1] corresponds to the highpass freq because that's the resolution
        avg_power_in_freq_band = np.mean(psd[:,1:high_idx], axis=-1)
        logMUA_signal[i] = np.squeeze(np.log(avg_power_in_freq_band/avg_power))

    new_signals = np.empty((len(subsample_times), channel_num))
    new_signals.fill(np.nan)
    new_signals[:, non_nan_channels] = logMUA_signal

    logMUA_asig = asig.duplicate_with_new_data(new_signals)
    logMUA_asig.array_annotations = asig.array_annotations
    logMUA_asig.sampling_rate = eff_logMUA_rate
    logMUA_asig.annotate(freq_band = [highpass_frequency, lowpass_frequency],
                         psd_frequency_resolution = highpass_frequency,
                         psd_overlap = psd_overlap,
                         psd_fs = fs)
    return logMUA_asig


def plot_logMUA_estimation(asig, logMUA_asig, highpass_frequency, lowpass_frequency,
                           t_start, t_stop, channel):
    asig = time_slice(asig, t_start, t_stop)
    logMUA_asig = time_slice(logMUA_asig, t_start, t_stop)
    filt_asig = butter(asig, highpass_frequency=highpass_frequency,
                             lowpass_frequency=lowpass_frequency)

    sns.set(style='ticks', palette="deep", context="notebook")
    fig, ax = plt.subplots()

    ax.plot(asig.times,
            zscore(asig.as_array()[:,channel]),
            label='original signal')

    ax.plot(filt_asig.times,
            zscore(filt_asig.as_array()[:,channel]) + 10,
            label=f'signal [{highpass_frequency}-{lowpass_frequency}]',
            alpha=0.5)

    ax.plot(logMUA_asig.times,
            zscore(logMUA_asig.as_array()[:,channel]) + 20,
            label='logMUA')

    ax.set_title('Channel {}'.format(channel))
    ax.set_xlabel('time [{}]'.format(asig.times.units.dimensionality.string))
    ax.set_yticklabels([])
    plt.legend()
    return ax


if __name__ == '__main__':
    args, unknown = CLI.parse_known_args()

    block = load_neo(args.data)

    logMUA_rate = None if args.logMUA_rate is None \
                  else args.logMUA_rate*pq.Hz

    fft_slice = None if args.fft_slice is None \
                else args.fft_slice*pq.s

    asig = logMUA_estimation(block.segments[0].analogsignals[0],
                             highpass_frequency=args.highpass_frequency*pq.Hz,
                             lowpass_frequency=args.lowpass_frequency*pq.Hz,
                             logMUA_rate=logMUA_rate,
                             psd_overlap=args.psd_overlap,
                             fft_slice=fft_slice)

    if args.plot_channels is not None:
        for channel in args.plot_channels:
            plot_logMUA_estimation(asig=block.segments[0].analogsignals[0],
                                   logMUA_asig=asig,
                                   highpass_frequency=args.highpass_frequency*pq.Hz,
                                   lowpass_frequency=args.lowpass_frequency*pq.Hz,
                                   t_start=args.plot_tstart,
                                   t_stop=args.plot_tstop,
                                   channel=channel)
            output_path = args.img_dir \
                        / args.img_name.replace('_channel0', f'_channel{channel}')
            save_plot(output_path)

    asig.description += "Estimated logMUA signal [{}, {}] Hz ({}). "\
                        .format(args.highpass_frequency, args.lowpass_frequency,
                                os.path.basename(__file__))
    block.segments[0].analogsignals[0] = asig

    write_neo(args.output, block)



"""
Calculate the wave directions by either interpolating trigger times and 
locations or by averaging the corresponding optical flow values.
"""

import neo
import numpy as np
import quantities as pq
import matplotlib.pyplot as plt
from matplotlib import patches
import os
import warnings
import argparse
import scipy
import pandas as pd
import seaborn as sns
from utils.io_utils import load_neo, save_plot
from utils.parse import none_or_str

CLI = argparse.ArgumentParser()
CLI.add_argument("--data", nargs='?', type=str, required=True,
                 help="path to input data in neo format")
CLI.add_argument("--output", nargs='?', type=str, required=True,
                 help="path of output file")
CLI.add_argument("--output_img", nargs='?', type=none_or_str, default=None,
                 help="path of output image file")
CLI.add_argument("--method", "--DIRECTION_METHOD", nargs='?', type=str, default='trigger_interpolation',
                 help="'tigger_interpolation' or 'optical_flow'")
CLI.add_argument("--event_name", "--EVENT_NAME", nargs='?', type=str, default='wavefronts',
                 help="name of neo.Event to analyze (must contain waves)")
                    
def calc_displacement(times, locations):
    slope, offset, _, _, stderr = scipy.stats.linregress(times, locations)
    d0, d1 = offset + slope*times[0], offset + slope*times[-1]
    displacement = d1 - d0
    displacement_err = np.sqrt(stderr**2 + (stderr*(times[-1]-times[0]))**2)
    return displacement, displacement_err

def trigger_interpolation(evts):
    spatial_scale = evts.annotations['spatial_scale']
    wave_ids = np.unique(evts.labels)
    dx_avg, dy_avg = np.zeros(len(wave_ids)), np.zeros(len(wave_ids))
    dx_std, dy_std = np.zeros(len(wave_ids)), np.zeros(len(wave_ids))

    # loop over waves
    for i, wave_i in enumerate(wave_ids):
        # Fit wave displacement
        idx = np.where(evts.labels == wave_i)[0]
        dx_avg[i], dx_std[i] = calc_displacement(evts.times[idx].magnitude,
                                   evts.array_annotations['x_coords'][idx]
                                 * spatial_scale.magnitude)
        dy_avg[i], dy_std[i] = calc_displacement(evts.times[idx].magnitude,
                                   evts.array_annotations['y_coords'][idx]
                                 * spatial_scale.magnitude)
    return dx_avg, dy_avg, dx_std, dy_std


def times2ids(time_array, times_selection):
    return np.array([np.argmax(time_array>=t) for t in times_selection])

def calc_flow_direction(evts, asig):
    wave_ids = np.unique(evts.labels.astype(int))
    dx_avg, dy_avg = np.zeros(len(wave_ids)), np.zeros(len(wave_ids))
    dx_std, dy_std = np.zeros(len(wave_ids)), np.zeros(len(wave_ids))
    signals = asig.as_array()

    for i, wave_i in enumerate(wave_ids):
        idx = np.where(evts.labels.astype(int) == wave_i)[0]
        t_idx = times2ids(asig.times, evts.times[idx])
        channels = evts.array_annotations['channels'][idx]
        flow_vectors = signals[t_idx, channels]
        flow_vectors /= np.abs(flow_vectors)
        if np.isnan(flow_vectors).any():
            warnings.warn("Signals at trigger points contain nans!")
        dx_avg[i] = np.nanmean(np.real(flow_vectors))
        dx_std[i] = np.nanstd(np.real(flow_vectors))
        dy_avg[i] = np.nanmean(np.imag(flow_vectors))
        dy_std[i] = np.nanstd(np.imag(flow_vectors))
    return dx_avg, dy_avg, dx_std, dy_std

def plot_directions(dataframe, wave_ids,
                    orientation_top=None, orientation_right=None):
    directions = dataframe.direction_x + dataframe.direction_y*1j
    directions_std = dataframe.direction_x_std + dataframe.direction_y_std*1j

    ncols = int(np.round(np.sqrt(len(wave_ids)+1)))
    nrows = int(np.ceil((len(wave_ids)+1)/ncols))
    fig, ax = plt.subplots(nrows=nrows, ncols=ncols,
                           figsize=(3*nrows, 3*ncols))

    rmax = np.max(np.abs(directions))
    for i, (d, d_std) in enumerate(zip(directions, directions_std)):
        row = int(i/ncols)
        if ncols == 1:
            cax = ax[row]
        else:
            col = i % ncols
            cax = ax[row][col]

        cax.plot([0,np.real(d)], [0,np.imag(d)], color='r', alpha=0.8)
        ellipsis = patches.Ellipse(xy=(np.real(d), np.imag(d)),
                                   width=2*np.real(d_std), height=2*np.imag(d_std),
                                   alpha=0.5)
        cax.add_artist(ellipsis)
        cax.set_title('wave {}'.format(wave_ids[i]))
        if np.isfinite(rmax):
            cax.set_ylim((-rmax,rmax))
            cax.set_xlim((-rmax,rmax))
        cax.axhline(0, c='k')
        cax.axvline(0, c='k')
        # cax.axes.get_xaxis().set_visible(False)
        # cax.axes.get_yaxis().set_visible(False)
        cax.set_xticks([])
        cax.set_yticks([])

    if ncols == 1:
        cax = ax[-1]
    else:
        cax = ax[-1][-1]
        for i in range(len(directions), nrows*ncols):
            row = int(i/ncols)
            col = i % ncols
            ax[row][col].set_axis_off()

    cax.axhline(0, c='k')
    cax.axvline(0, c='k')
    cax.set_xlim((-2,2))
    cax.set_ylim((-2,2))
    if orientation_top is not None:
        cax.text(0, 1,orientation_top, rotation='vertical',
                        verticalalignment='center', horizontalalignment='right')
    if orientation_right is not None:
        cax.text(1, 0, orientation_right,
                        verticalalignment='top', horizontalalignment='center')

    sns.despine(left=True, bottom=True)
    return ax

if __name__ == '__main__':
    args, unknown = CLI.parse_known_args()

    block = load_neo(args.data)

    if args.method == 'optical_flow':
        if args.event_name == 'wavemodes':
            warnings.warn('The planar direction of wavemodes can not be '
                          'calculated with the optical_flow method. '
                          'Using trigger_interpolation instead.')
            args.method = 'trigger_interpolation'
        elif not len(block.filter(name='optical_flow', objects="AnalogSignal")):
            warnings.warn('No optical_flow signal could be found for the '
                          'calculation of planar directions. '
                          'Using trigger_interpolation instead.')
            args.method = 'trigger_interpolation'

    evts = block.filter(name=args.event_name, objects="Event")[0]
    evts = evts[evts.labels.astype('str') != '-1']

    if args.method == 'trigger_interpolation':
        dx_avg, dy_avg, dx_std, dy_std = trigger_interpolation(evts)
    elif args.method == 'optical_flow':
        asig = block.filter(name='optical_flow', objects="AnalogSignal")[0]
        dx_avg, dy_avg, dx_std, dy_std = calc_flow_direction(evts, asig)
    else:
        raise NameError(f'Method name {args.method} is not recognized!')

    df = pd.DataFrame(np.unique(evts.labels.astype(int)),
                      columns=[f'{args.event_name}_id'])
    df['direction_x'] = dx_avg
    df['direction_y'] = dy_avg
    df['direction_x_std'] = dx_std
    df['direction_y_std'] = dy_std

    if args.output_img is not None:
        plot_directions(df,
                        wave_ids=np.unique(evts.labels.astype(int)),
                        orientation_top=evts.annotations['orientation_top'],
                        orientation_right=evts.annotations['orientation_right'])
        save_plot(args.output_img)

    df.to_csv(args.output)
